# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.plugin.control_cli_plugin.control_service_configuration import (
    ControlServiceConfiguration,
)
from vdk.plugin.control_cli_plugin.control_service_properties_client import (
    ControlServicePropertiesServiceClient,
)

log = logging.getLogger(__name__)


@hookimpl
def initialize_job(context: JobContext) -> None:
    conf = ControlServiceConfiguration(context.core_context.configuration)
    url = conf.control_service_rest_api_url()
    if url:
        log.info("Initialize Control Service based Properties client implementation.")
        context.properties.set_properties_factory_method(
            "default", lambda: ControlServicePropertiesServiceClient(url)
        )
        context.properties.set_properties_factory_method(
            "control-service", lambda: ControlServicePropertiesServiceClient(url)
        )
    else:
        log.info(
            "Control Service REST API URL is not configured. "
            "Will not initialize Control Service based Properties client implementation."
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import os

import click
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.control.command_groups.common_group.default import (
    reset_default_command,
)
from vdk.internal.control.command_groups.common_group.default import set_default_command
from vdk.internal.control.command_groups.job.create import create
from vdk.internal.control.command_groups.job.delete import delete
from vdk.internal.control.command_groups.job.deploy_cli import deploy
from vdk.internal.control.command_groups.job.download_job import download_job
from vdk.internal.control.command_groups.job.download_key import download_key
from vdk.internal.control.command_groups.job.execute import execute
from vdk.internal.control.command_groups.job.list import list_command
from vdk.internal.control.command_groups.job.properties import properties_command
from vdk.internal.control.command_groups.job.show import show_command
from vdk.internal.control.command_groups.login_group.login import login
from vdk.internal.control.command_groups.logout_group.logout import logout
from vdk.internal.control.configuration.default_options import DefaultOptions
from vdk.internal.control.plugin import control_plugin_manager
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.context import CoreContext
from vdk.plugin.control_cli_plugin import control_service_configuration


@hookimpl(
    tryfirst=True
)  # this hook will be executed first so that these commands can be decorated later if needed
def vdk_command_line(root_command: click.Group):
    root_command.add_command(login)
    root_command.add_command(logout)
    root_command.add_command(delete)
    root_command.add_command(create)
    root_command.add_command(download_key)
    root_command.add_command(list_command)
    root_command.add_command(deploy)
    root_command.add_command(execute)
    root_command.add_command(download_job)
    root_command.add_command(set_default_command)
    root_command.add_command(reset_default_command)
    root_command.add_command(show_command)
    root_command.add_command(properties_command)

    plugins = control_plugin_manager.Plugins()
    default_options = DefaultOptions(plugins)
    if default_options.get_default_map():
        root_command.context_settings["default_map"] = default_options.get_default_map()


@hookimpl(tryfirst=True)
def vdk_configure(config_builder: ConfigurationBuilder) -> None:
    control_service_configuration.add_definitions(config_builder)


def __set_value_as_env_var(context: CoreContext, config_key: str):
    env_var_name = f"VDK_{config_key.upper()}"
    if env_var_name not in os.environ and context.configuration.get_value(config_key):
        os.environ[env_var_name] = str(context.configuration.get_value(config_key))


@hookimpl
def vdk_initialize(context: CoreContext) -> None:
    """Update vdk-control-cli configuration.

    vdk-control-cli is not a proper plugin yet and looks only at env
    variables for configuration. so we are setting them here.
    """
    __set_value_as_env_var(context, "API_TOKEN")
    __set_value_as_env_var(context, "API_TOKEN_AUTHORIZATION_URL")
    __set_value_as_env_var(context, "CONTROL_SERVICE_REST_API_URL")
    __set_value_as_env_var(context, "CONTROL_SAMPLE_JOB_DIRECTORY")
    __set_value_as_env_var(context, "CONTROL_HTTP_VERIFY_SSL")
    __set_value_as_env_var(context, "CONTROL_HTTP_TOTAL_RETRIES")
    __set_value_as_env_var(context, "CONTROL_HTTP_READ_RETRIES")
    __set_value_as_env_var(context, "CONTROL_HTTP_READ_TIMEOUT_SECONDS")
    __set_value_as_env_var(context, "CONTROL_HTTP_CONNECT_RETRIES")
    __set_value_as_env_var(context, "CONTROL_HTTP_CONNECT_TIMEOUT_SECONDS")



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
from typing import List

from taurus_datajob_api import DataJobExecution
from vdk.api.plugin.hook_markers import hookimpl
from vdk.internal.builtin_plugins.config import vdk_config
from vdk.internal.builtin_plugins.config.job_config import JobConfigKeys
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.builtin_plugins.termination_message.writer import (
    TerminationMessageWriterPlugin,
)
from vdk.internal.control.configuration.vdk_config import VDKConfig
from vdk.internal.control.rest_lib.factory import ApiClientFactory
from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.statestore import CommonStoreKeys

log = logging.getLogger(__name__)
EXECUTION_SKIP_CHECKER_ENABLED = "EXECUTION_SKIP_CHECKER_ENABLED"


class ConcurrentExecutionChecker:
    """
    The class make sure to check if there is another execution of the data job already started.
    It is used to prevent concurrent executions from getting started in the same time.
    As often data jobs are not expected to be parallelizable since they do maintain state - for example
    ingestion job that pulls data since last time it ran and moves it to a warehouse would duplicate it
    if started concurrently.
    """

    def __init__(self, rest_api_url: str) -> None:
        # set as protected instead of private so it can be mocked in unit tests
        self._execution_api_client = ApiClientFactory(rest_api_url).get_execution_api()

    def is_job_execution_running(
        self, job_name, job_team, job_execution_attempt_id
    ) -> bool:
        """
        :param job_name : str value for the job name
        :param job_team : str value for the job team
        :param job_execution_attempt_id : str current job's execution id as returned by VDK execution_state
        This method calls the "/data-jobs/for-team/{team_name}/name/{job_name}/executions"
        endpoint with the additional params "submitted" and "running". The API will return all jobs that are
        submitted and running with the provided team and job name. We have the following possible scenarios:
        API returns an empty list - this means no other data jobs excecutions are running and we can proceed with current execution.
        API returns one execution with a different ID - this means that another excecution is running and we should skip.
        API returns one job with the same ID - this means that we can proceed with excecution.
        API returns two jobs one of which has a different ID - we need to skip current excecution.
        It then parses the response and returns a boolean.
        :return: bool
                 If the data job has a different execution running in K8S.
                 A running different execution means that there is a data job execution currently
                 submitted or running with a different id and in that case we should skip current execution
                 until the other finishes.
        """
        jobs = self._get_running_job_executions(job_name, job_team)
        job_execution_running = False
        for job_execution in jobs:
            # We break only if we have a job with a different execution id
            # Sometimes ids returned by VDK execution_state can slightly differ than
            # ids returned by the API. VDK ids  have '-xxxxx' appended at the end,
            # where x is a random alphameric char. Otherwise ids start the same.
            if not job_execution_attempt_id.startswith(job_execution.id):
                log.info(
                    f"Found another running job execution for {job_name}. Current execution id: {job_execution_attempt_id}. Other running execution id: {job_execution.id}"
                )
                job_execution_running = True
                break

        return job_execution_running

    def _get_running_job_executions(
        self, job_name, team_name
    ) -> List[DataJobExecution]:
        job_status = [
            "submitted",
            "running",
        ]  # We query only for running or manual jobs
        response = self._execution_api_client.data_job_execution_list(
            job_name=job_name, team_name=team_name, execution_status=job_status
        )
        log.info(f"Api call returned job execution list: {response}")
        return response


def _skip_job_run(job_name) -> None:
    log.info(f"Data job: {job_name} is already running in K8S.")
    log.info(
        f"Skipping execution and exiting. You can retry the job when it finishes executing in K8S."
    )
    os._exit(0)


def _skip_job_if_necessary(
    log_config: str,
    job_name: str,
    execution_id: str,
    job_team: str,
    configuration: Configuration,
):
    try:
        log.info("Checking if job should be skipped:")
        log.info(
            f"Job : {job_name}, Team : {job_team}, Log config: {log_config}, execution_id: {execution_id}"
        )
        # TODO: Do not use log config type to check if it is a cloud run or not.
        if log_config != "CLOUD":
            logging.info("Local execution, skipping parallel execution check.")
            return None

        vdk_cfg = VDKConfig()
        job_checker = ConcurrentExecutionChecker(vdk_cfg.control_service_rest_api_url)
        job_running = job_checker.is_job_execution_running(
            job_name, job_team, execution_id
        )

        if job_running:
            log.info(f"Skipping job {job_name}")
            writer_plugin = TerminationMessageWriterPlugin()
            writer_plugin.write_termination_message(
                configuration=configuration, execution_skipped=True
            )
            _skip_job_run(job_name)  # calls os._exit(0)
            return 1  # All other branches return None
    except Exception as exc:
        log.warning(
            f"Error while checking for another data job excecution: {str(exc)} "
        )
        log.warning("Printing stack trace and proceeding with execution:")
    return None


@hookimpl(tryfirst=True)
def vdk_configure(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=EXECUTION_SKIP_CHECKER_ENABLED,
        default_value=True,
        description="Set to false if you want to disable execution skipping logic entirely.",
    )


@hookimpl(tryfirst=True)
def run_job(context: JobContext) -> None:
    try:
        configuration = context.core_context.configuration

        if not configuration.get_value(EXECUTION_SKIP_CHECKER_ENABLED):
            log.info(
                "Skipping logic is turned off. Continuing with current execution without checking for another running execution."
            )
            return None

        log_config_type = configuration.get_value(vdk_config.LOG_CONFIG)
        job_name = context.name
        job_team = configuration.get_value(JobConfigKeys.TEAM)
        execution_id = context.core_context.state.get(CommonStoreKeys.EXECUTION_ID)

        return _skip_job_if_necessary(
            log_config_type, job_name, execution_id, job_team, configuration
        )
    except Exception as exc:
        log.warning(f"Error while setting up arguments: {str(exc)}")
        log.warning("Catching exception and continuing with execution:")
        log.warning(exc)
        return None



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.internal.core.config import Configuration
from vdk.internal.core.config import ConfigurationBuilder

API_TOKEN = "API_TOKEN"
API_TOKEN_AUTHORIZATION_URL = "API_TOKEN_AUTHORIZATION_URL"
CONTROL_SERVICE_REST_API_URL = "CONTROL_SERVICE_REST_API_URL"
CONTROL_SAMPLE_JOB_DIRECTORY = "CONTROL_SAMPLE_JOB_DIRECTORY"
CONTROL_HTTP_VERIFY_SSL = "CONTROL_HTTP_VERIFY_SSL"
CONTROL_HTTP_TOTAL_RETRIES = "CONTROL_HTTP_TOTAL_RETRIES"
CONTROL_HTTP_READ_RETRIES = "CONTROL_HTTP_READ_RETRIES"
CONTROL_HTTP_READ_TIMEOUT_SECONDS = "CONTROL_HTTP_READ_TIMEOUT_SECONDS"
CONTROL_HTTP_CONNECT_RETRIES = "CONTROL_HTTP_CONNECT_RETRIES"
CONTROL_HTTP_CONNECT_TIMEOUT_SECONDS = "CONTROL_HTTP_CONNECT_TIMEOUT_SECONDS"


class ControlServiceConfiguration:
    def __init__(self, config: Configuration) -> None:
        self.__config = config

    def api_token(self):
        return self.__config.get_value(API_TOKEN)

    def api_token_authorization_url(self):
        return self.__config.get_value(API_TOKEN_AUTHORIZATION_URL)

    def control_service_rest_api_url(self):
        return self.__config.get_value(CONTROL_SERVICE_REST_API_URL)

    def control_sample_job_directory(self):
        return self.__config.get_value(CONTROL_SAMPLE_JOB_DIRECTORY)

    def control_http_verify_ssl(self):
        return self.__config.get_value(CONTROL_HTTP_VERIFY_SSL)

    def control_http_total_retries(self):
        return self.__config.get_value(CONTROL_HTTP_TOTAL_RETRIES)

    def control_http_read_retries(self):
        return self.__config.get_value(CONTROL_HTTP_READ_RETRIES)

    def control_http_read_timeout_seconds(self):
        return self.__config.get_value(CONTROL_HTTP_READ_TIMEOUT_SECONDS)

    def control_http_connect_retries(self):
        return self.__config.get_value(CONTROL_HTTP_CONNECT_RETRIES)

    def control_http_connect_timeout_seconds(self):
        return self.__config.get_value(CONTROL_HTTP_CONNECT_TIMEOUT_SECONDS)


def add_definitions(config_builder: ConfigurationBuilder):
    config_builder.add(
        key=API_TOKEN,
        default_value=None,
        description="API Token for the OAuth2 provider used in exchange for Access Token "
        "which is then used in any request to Control Service. "
        "It is acts as 'refresh token' (https://datatracker.ietf.org/doc/html/rfc6749#section-1.5)"
        " - in other words it's used as credentials used to obtain access tokens. "
        "That request is against URL specified by API_TOKEN_AUTHORIZATION_URL. "
        "Default API Token used if another authentication has not been used with vdk login.",
    )
    config_builder.add(
        key=API_TOKEN_AUTHORIZATION_URL,
        default_value=None,
        description="Location of the API Token OAuth2 provider. "
        "It is used to exchange api refresh token for access token. "
        "See https://tools.ietf.org/html/rfc6749#section-5.1"
        "It is used alongside API_TOKEN if vdk login has not been run.",
    )
    config_builder.add(
        key=CONTROL_SERVICE_REST_API_URL,
        default_value=None,
        description="The default base REST API URL. It looks like http://server (without path e.g. /data-jobs).",
    )
    config_builder.add(
        key=CONTROL_SAMPLE_JOB_DIRECTORY,
        default_value=None,
        description="Directory used to create sample job locally during vdk create --local operation. "
        "If not set it will use a predefined prepared one that should show basic sample job.",
    )
    config_builder.add(
        key=CONTROL_HTTP_VERIFY_SSL,
        default_value=True,
        description="Verify SSL certificate of Control Service Server.",
    )
    config_builder.add(
        key=CONTROL_HTTP_TOTAL_RETRIES,
        default_value=None,
        description="Total number of http retries to allow for Control Service API requests.",
    )
    config_builder.add(
        key=CONTROL_HTTP_READ_RETRIES,
        default_value=None,
        description="Total number of http retries to allow for Control Service API requests.",
    )
    config_builder.add(
        key=CONTROL_HTTP_CONNECT_RETRIES,
        default_value=None,
        description="How many connection-related errors to retry on against Control Service API Server."
        "These are errors raised before the request is sent to the remote server, "
        "which we assume has not triggered the server to process the request.",
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import pathlib

import setuptools

"""
Builds a package with the help of setuptools in order for this package to be imported in other projects
"""

__version__ = "0.1.477708478"

setuptools.setup(
    name="vdk-plugin-control-cli",
    version=__version__,
    url="https://github.com/vmware/versatile-data-kit",
    description="Versatile Data Kit SDK plugin exposing CLI commands for managing the lifecycle of a Data Jobs.",
    long_description=pathlib.Path("README.md").read_text(),
    long_description_content_type="text/markdown",
    install_requires=["vdk-core", "vdk-control-cli", "requests"],
    package_dir={"": "src"},
    packages=setuptools.find_namespace_packages(where="src"),
    entry_points={
        "vdk.plugin.run": [
            "vdk-plugin-control-cli = vdk.plugin.control_cli_plugin.vdk_plugin_control_cli",
            "vdk-control-service-properties = vdk.plugin.control_cli_plugin.properties_plugin",
            "vdk-execution-skip = vdk.plugin.control_cli_plugin.execution_skip",
        ]
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import functools
import json
import logging

from taurus_datajob_api import ApiException
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import PlatformServiceError
from vdk.internal.core.errors import UserCodeError
from vdk.internal.core.errors import VdkConfigurationError

log = logging.getLogger(__name__)


class ConstrolServiceApiErrorDecorator:
    """Decorator for exceptions raised by Control Service API.

    It tries to present them in user friendly way and detect the
    resolved by type of the error.
    """

    def __init__(
        self, what="Control Service Error", consequences="Operation cannot complete."
    ):
        """
        :param what: what is the error message (see VDK error standard). Will be set if exception does not provide one.
        :param consequences: what are the consequences for the user (see VDK error standard).
        Will be set if exception does not provide one.
        """
        self.__what = what
        self.__consequences = consequences

    def _get_error(self, exception: ApiException):
        try:
            return json.loads(exception.body)
        except Exception as jsonException:
            log.debug(
                f"We failed to parse API exception body as JSON due to {jsonException}. "
                f"Will fallback to use generic template for error message."
            )
            error = {}
            if exception.status == 401:
                error = dict(
                    what=self.__what,
                    why=f"The request has not been applied because it lacks valid authentication credentials. "
                    f"Error returned by control service is {exception.body}",
                    consequences=self.__consequences,
                    countermeasures="Try to login again using VDK CLI login command."
                    "Or set correct api token configuration (see vdk config-help)."
                    "Make sure you have permission to execute the given operation."
                    "Investigate (google) the error returned by control service."
                    "And if all fails try to contact the support team.",
                )
            elif exception.status == 404:
                error = dict(
                    what=self.__what,
                    why="The requested resource cannot be found. "
                    "You may have a spelling mistake or the data job has not been created",
                    consequences=self.__consequences,
                    countermeasures="Make sure that the data job name and team name are spelled correctly "
                    "(it is case-sensitive) . "
                    "Make sure you have run `vdk create --cloud` before doing this operation on a data job.",
                )
            else:
                error = dict(
                    what=error.get("what", self.__what),
                    why=error.get(
                        "why",
                        f"Http error: status: {exception.status} - {exception.body}",
                    ),
                    consequence=error.get("consequences", self.__consequences),
                    countermeasure=error.get(
                        "countermeasures",
                        "See error and try to resolve it or open ticket to SRE team.",
                    ),
                )

            return error

    def __call__(self, fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            try:
                result = fn(*args, **kwargs)
                return result
            except ApiException as ex:
                log.debug(
                    f"An API Exception occurred in {fn.__module__}.{fn.__name__}",
                    f"The Exception class was: {ex}",
                )

                body = self._get_error(ex)
                if ex.status == 401:
                    raise VdkConfigurationError(body) from ex
                elif 400 <= ex.status < 500:
                    raise UserCodeError(body) from ex
                elif ex.status >= 500:
                    raise PlatformServiceError(body) from ex
            except Exception as ex:
                error = ErrorMessage(
                    summary="",
                    what=self.__what,
                    why=f"Error: {ex}",
                    consequences=self.__consequences,
                    countermeasures="See error and try to resolve it or open ticket to SRE team.",
                )
                raise PlatformServiceError(error) from ex

        return decorated



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from typing import Dict

from vdk.api.plugin.plugin_input import IPropertiesServiceClient
from vdk.internal.control.rest_lib.factory import ApiClientFactory
from vdk.internal.control.rest_lib.rest_client_errors import ApiClientErrorDecorator
from vdk.plugin.control_cli_plugin.control_service_api_error_decorator import (
    ConstrolServiceApiErrorDecorator,
)

log = logging.getLogger(__name__)


class ControlServicePropertiesServiceClient(IPropertiesServiceClient):
    """Implementation of PropertiesServiceClient which connects to VDK Control
    Service Properties API."""

    def __init__(self, rest_api_url: str):
        self.__properties_api = ApiClientFactory(rest_api_url).get_properties_api()
        log.debug(f"Initialized Properties against {rest_api_url}.")

    @ConstrolServiceApiErrorDecorator()
    def read_properties(self, job_name: str, team_name: str):
        data = self.__properties_api.data_job_properties_read(
            team_name=team_name, job_name=job_name, deployment_id="TODO"
        )
        return data

    @ConstrolServiceApiErrorDecorator()
    def write_properties(self, job_name: str, team_name: str, properties: Dict):
        self.__properties_api.data_job_properties_update(
            team_name=team_name,
            job_name=job_name,
            deployment_id="TODO",
            request_body=properties,
        )
